classdef contacts < handle
    % forces on car-body from springs / dampeners
    % as a function of y 
    % for front and rear axle
    
    properties
        a
        kT
        s
        h
        eps
        T_eps
        T_s
        T_h
        R
        vRef

        springC
        trackC
        coeffsA
        mu
        mu0
        muC
        muEps
        section
        contactCoeffs
        lastq
        localTrack
        sinphi
        cosphi
        gamma
        singamma
        cosgamma
        QB   % battery potential
        UB2  % voltage over capacity
        TB   % temperature
        u    % coordinates
        U    % absolute wheel coordinates
        v    % velocities
        V    % absolute wheel velocities in refernce coorddinates
        omega% wheel speed
        phi  % body-rotation
        phiP % body angular velocity
        E    % integral value of error value e
        IM   % motor current
        
        UB   % battery voltage -> to be computed from state variables in "batteries"
        
        vC   % contact speeds ...
        vT0  % loss velocity
        eta  % relative intrusion
    end
    
    methods
        function obj = contacts(car)
            % constructor class
            % read hashMap
            obj.kT   = car('k[T]');
            obj.eps  = car('eps'); 
            obj.R    = car('R');
            obj.a    = [car('a[1]'),car('a[2]'),car('a[3]')]; % drop a[0]
            obj.vRef = car('v[ref]');
            %obj.T_s  = car('T_s');
            %obj.T_h  = car('T_h');
            %obj.T_eps= car('T_eps');
            obj.mu   = car('mu');
            obj.mu0  = car('mu0');
            obj.muEps= car('mu_eps');
            obj.vT0  = car('v[T0]');
            obj.section = [1,1]; % section where axles had last contact with track (start with "1")
            obj.localTrack = []; %linear polynomial of track at contact)
            obj.vC = zeros(2,2);
            obj.singamma = zeros(1,2);
            obj.cosgamma = zeros(1,2);
            obj.gamma = zeros(1,2);
            obj.eta = zeros(1,2);
        end

        function initialize(obj, track)
            % compute missing properties
            
            % collect arrays s, h and eps from track.keys 
            % example: s[1]=track('s[1]')
            sectionId = 1;
            while true
                % obs: start with "0"
                objName = ['s[',num2str(sectionId-1),']'];
                if isKey(track,objName)
                    obj.T_s(sectionId) = track(objName);
                else
                    break;
                end
                objName = ['h[',num2str(sectionId-1),']'];
                if isKey(track, objName)
                    obj.T_h(sectionId) = track(objName);
                end
                objName = ['eps[',num2str(sectionId),']'];
                if isKey(track,objName)
                    obj.T_eps(sectionId) = track(objName);
                end
                % next section
                sectionId = sectionId+1;
            end
            
            % define characteristic

            obj.trackC  = characteristics(obj.T_s,obj.T_h,obj.T_eps,'Track','s/L->','h/L'); 
            % check quality (can the wheel get stuck in a dale ....)
            for round = 1:size(obj.T_eps,1)
                secId = 2*round;
                % parable is open to the ceiling -> builds a trough
                if obj.trackC.poly(secId,1)>0 && obj.T_eps(round) < obj.R
                    warning("tracks rounding radius (eps = %f) is smaller than wheel radius R = %f in troughs. This will lead to errors.",obj.T_eps(round),obj.R);
                end
            end
            % ...............
            obj.springC = characteristics([-1 0 1],[0 0 obj.kT*1],[10.*obj.eps],'Track Spring','Dc/L->','F/(M*L/T^2)->');                     
            obj.muC     = characteristics([-1 -0.2 -2*obj.muEps +2*obj.muEps +0.2 +1],...
                                        [-obj.mu -obj.mu -obj.mu0 +obj.mu0 +obj.mu +obj.mu],...
                                        [0.1 obj.muEps obj.muEps 0.1],...
                                          'Friction','v/(L/T)','mu/1');

            % initial guess q0 / last results q from Lagragian Minimization
            obj.lastq = zeros(2,4,obj.trackC.nos);
            for secId = 2:2:obj.trackC.nos
                X = obj.boundsOfSection(secId);
                obj.lastq (1,:,secId)= [(X(2)+X(1))/2, 1., 0, 1.];
            end                                         
                                      
            % compute coefficients for solution of [xi/eta] for the plane surfaces   
            % xi  = A[1,2]*u[3]+A[1,3]*u[4]+A[1,1], 
            % eta = A[2,2]*u[3]+A[2,3]*u[4]+A[2,1]
            % from contact.mac
            %A[ 1 , 1 ]= Δs/(Δs^2+Δh^2) 
            %A[ 1 , 2 ]= Δh/(Δs^2+Δh^2) 
            %A[ 2 , 1 ]= -(Δh^2*sqrt(Δs^2/Δh^2+1))/(R*Δs^2+R*Δh^2) 
            %A[ 2 , 2 ]= (Δh*Δs*sqrt(Δs^2/Δh^2+1))/(R*Δs^2+R*Δh^2) 
            %b[ 1 ]= -(s[i]*Δs)/(Δs^2+Δh^2)-(h[i]*Δh)/(Δs^2+Δh^2) 
            %b[ 2 ]= (s[i]*Δh^2*sqrt(Δs^2/Δh^2+1))/(R*Δs^2+R*Δh^2)-(h[i]*Δh*Δs*sqrt(Δs^2/Δh^2+1))/(R*Δs^2+R*Δh^2)           

            obj.coeffsA = [];
            s = obj.trackC.x; %#ok<*PROP>
            h = obj.trackC.y; %#ok<*PROP>
            R = obj.R;
            for i=1:obj.trackC.nos
                obj.coeffsA = [obj.coeffsA;zeros(2,3)];
                if rem(i,2) ~= 0
                    % the result [xi,eta] = A[:,2:3].[u[3],u[4]]+A[:,1]
                    % coeffs of u[3], u[4]
                    Ds = s(i+1)-s(i);
                    Dh = h(i+1)-h(i);
                    % 
                    obj.coeffsA(2*i-1,2) = Ds/(Ds^2+Dh^2);
                    obj.coeffsA(2*i-1,3) = Dh/(Ds^2+Dh^2);
                    % constant part
                    obj.coeffsA(2*i-1,1) = -(s(i)*Ds)/(Ds^2+Dh^2)-(h(i)*Dh)/(Ds^2+Dh^2);
                    if Dh ~= 0
                        obj.coeffsA(2*i  ,2) = -(Dh^2*sqrt(Ds^2/Dh^2+1))/(R*Ds^2+R*Dh^2);
                        obj.coeffsA(2*i  ,3) =  (Dh*Ds*sqrt(Ds^2/Dh^2+1))/(R*Ds^2+R*Dh^2);
                        obj.coeffsA(2*i  ,1) =  (s(i)*Dh^2*sqrt(Ds^2/Dh^2+1))/(R*Ds^2+R*Dh^2)-(h(i)*Dh*Ds*sqrt(Ds^2/Dh^2+1))/(R*Ds^2+R*Dh^2);
                    else
                        obj.coeffsA(2*i  ,2) =  0;
                        obj.coeffsA(2*i  ,3) =  1/R;
                        obj.coeffsA(2*i  ,1) = -h(i)/R;                        
                    end
                end
            % test
            % A = contacts.coeffsA(1:2,2:3)
            % b = contacts.coeffsA(1:2,1)
            % u = [[2];[2]]
            % xi = A*u + b
            end
        end

        function putY(obj, y)
            % saves y in the class
            %obj.u(1:4) = y( 7:10);
            %obj.phi    = y(11);
            %obj.v(1:4) = y( 1: 4);
            %obj.phiP   = y( 5);
            %obj.omega  = y( 6);  

            % battery
            obj.QB  = y(1);
            obj.UB2 = y(2);
            obj.TB  = y(3);
            
            % car body
            offset =3;             
            obj.v(1) = y(offset+ 1);
            obj.v(2) = y(offset+ 2);
            obj.v(3) = y(offset+ 3);
            obj.v(4) = y(offset+ 4);
            obj.phiP = y(offset+ 5);
            obj.omega= y(offset+ 6);  
            obj.u(1) = y(offset+ 7);
            obj.u(2) = y(offset+ 8);
            obj.u(3) = y(offset+ 9);
            obj.u(4) = y(offset+10);
            obj.phi  = y(offset+11);

            % driver controls
            obj.E   = y(15);
            
            % e-motor and drive-train
            obj.IM  = y(16);
            
            % trigonometric functions abbreviated
            obj.cosphi = cos(obj.phi);
            obj.sinphi = sin(obj.phi);

            % absolute position of wheel axles ( (U[3],U[4]) / (U[5],U[6]))
            obj.U(3) = -(obj.u(3)-obj.a(3))*obj.sinphi+obj.a(1)*obj.cosphi+obj.u(1);
            obj.U(4) = + obj.a(1)*obj.sinphi+(obj.u(3)-obj.a(3))*obj.cosphi+obj.u(2);
            obj.U(5) = -(obj.u(4)-obj.a(3))*obj.sinphi-obj.a(2)*obj.cosphi+obj.u(1);
            obj.U(6) = - obj.a(2)*obj.sinphi+(obj.u(4)-obj.a(3))*obj.cosphi+obj.u(2);

            % absolute velocities of wheel axles( (V[3],V[4]) / (V[5],V[6]))
            obj.V(3) = ((obj.a(3)-obj.u(3))*obj.cosphi-obj.a(1)*obj.sinphi)*obj.phiP-obj.v(3)*obj.sinphi+obj.v(1);
            obj.V(4) = ((obj.a(3)-obj.u(3))*obj.sinphi+obj.a(1)*obj.cosphi)*obj.phiP+obj.v(3)*obj.cosphi+obj.v(2);
            obj.V(5) = ( obj.a(2)*obj.sinphi+(obj.a(3)-obj.u(4))*obj.cosphi)*obj.phiP-obj.v(4)*obj.sinphi+obj.v(1);
            obj.V(6) = ((obj.a(3)-obj.u(4))*obj.sinphi-obj.a(2)*obj.cosphi)*obj.phiP+obj.v(4)*obj.cosphi+obj.v(2);

        end

        function xi = getContactGeometry(obj, axle)
            % compute intrusion in track DR and angle gamma
            
            assert(any([1,2]== axle), "Axle no must be '1' oder '2'");
            
            while true
                %disp("secion for axle "+num2str(axle)+": "+obj.section(1,axle));
                %pause
                % continue until you're in the right section
                if rem(obj.section(1,axle),2) ~= 0
                    % Track ist eine Gerade (ids 1, 3, 5, ....)
                    [xi,eta,gamma] = wheelInContactWithPlane(obj,axle);
                else
                    % Track is parable (ids 2, 4, 6, ....)
                    [xi,eta,gamma] = wheelInContactWithParable(obj,axle);
                end
                % disp('?????????????????');
                % disp(obj.section(1,axle));
                % disp(xi);
                % disp(eta);
                % disp(gamma);
                % disp('?????????????????');
                % need to swap section?
                if  0.<=xi && xi<=1.
                    break; % YES - DONE
                elseif xi>1.
                    % break;
                    if obj.section(1,axle)==obj.trackC.nos
                        break; % YES - DONE - there are no more sections
                    end
                    % axle runs in new section
                    %disp("xi = "+num2str(xi)+": section one up   for axle "+num2str(axle));
                    %pause;
                    obj.section(1,axle)=obj.section(1,axle)+1;
                elseif obj.section(1,axle)==1
                    % section 1 is extended towads - Infinity.
                    break; % YES - axle is left of first point
                else
                    %disp("xi = "+num2str(xi)+": section one down for axle "+num2str(axle));
                    %pause;
                    obj.section(1,axle)=obj.section(1,axle)-1;
                end
            end
            % eta
            obj.eta(axle) = eta;
            %
            obj.gamma(axle) = gamma;
            obj.singamma(axle) = sin(gamma);
            obj.cosgamma(axle) = cos(gamma);
        end
        
        function [xi,eta,gamma] = wheelInContactWithPlane(obj,axle)
            % 
            row = 2*obj.section(1,axle)-1;
            row = [row,row+1];
            % get coordinates
            % Achtung: großes U hier!
            u = obj.U(2+2*axle-1:2+2*axle); %#ok<PROPLC>
            r = obj.coeffsA(row,2:3)*u' + obj.coeffsA(row,1); %#ok<PROPLC>
            
            % get left and right end to comute x from xi
            xi  = r(1);
            X = boundsOfSection(obj, obj.section(1,axle));
            x = X(1)+xi*(X(2)-X(1));
            % take slant from polznomial
            gamma = atan(polyval(polyder(obj.trackC.poly(obj.section(1,axle),:)),x));
          
            % reverse eta for negative slant
            if gamma < 0
                eta = -r(2);
            else
                eta = +r(2);
            end
                        
        end
        
        function X = boundsOfSection(obj,secId)
            % get x-bounds of section
            X = [obj.trackC.x(secId),obj.trackC.x(secId+1)];
        end

        function [xi,eta,gamma] = wheelInContactWithParable(obj,axle)
            % use solution of nonlinear equation from Maxima
            secId = obj.section(1,axle);
            p0 = obj.trackC.poly(secId,3);
            p1 = obj.trackC.poly(secId,2);
            p2 = obj.trackC.poly(secId,1);

            U = [obj.U(2*axle+1),obj.U(2*axle+2)]; %#ok<PROPLC>
            
            % from contact.mac, Maxima 
            xSet = [-((sqrt(3)*i+1)*(sqrt((64*p0^3-192*U(2)*p0^2+192*U(2)^2*p0-64*U(2)^3)*p2^3+((-48*p0^2+96*U(2)*p0-48*U(2)^2)*p1^2+96*p0^2-192*U(2)*p0+96*U(2)^2+108*U(1)^2)*p2^2+((12*p0-12*U(2))*p1^4+(48*U(2)-48*p0)*p1^2+108*U(1)*p1+48*p0-48*U(2))*p2-p1^6+6*p1^4+15*p1^2+8)+2*3^(3/2)*U(1)*p2+3^(3/2)*p1)^(2/3)+2*sqrt(3)*p1*(sqrt((64*p0^3-192*U(2)*p0^2+192*U(2)^2*p0-64*U(2)^3)*p2^3+((-48*p0^2+96*U(2)*p0-48*U(2)^2)*p1^2+96*p0^2-192*U(2)*p0+96*U(2)^2+108*U(1)^2)*p2^2+((12*p0-12*U(2))*p1^4+(48*U(2)-48*p0)*p1^2+108*U(1)*p1+48*p0-48*U(2))*p2-p1^6+6*p1^4+15*p1^2+8)+2*3^(3/2)*U(1)*p2+3^(3/2)*p1)^(1/3)+((4*sqrt(3)*i-4)*p0+(4-4*sqrt(3)*i)*U(2))*p2+(1-sqrt(3)*i)*p1^2+2*sqrt(3)*i-2)/(4*sqrt(3)*p2*(sqrt((64*p0^3-192*U(2)*p0^2+192*U(2)^2*p0-64*U(2)^3)*p2^3+((-48*p0^2+96*U(2)*p0-48*U(2)^2)*p1^2+96*p0^2-192*U(2)*p0+96*U(2)^2+108*U(1)^2)*p2^2+((12*p0-12*U(2))*p1^4+(48*U(2)-48*p0)*p1^2+108*U(1)*p1+48*p0-48*U(2))*p2-p1^6+6*p1^4+15*p1^2+8)+2*3^(3/2)*U(1)*p2+3^(3/2)*p1)^(1/3));...
                  ((sqrt(3)*i-1)*(sqrt((64*p0^3-192*U(2)*p0^2+192*U(2)^2*p0-64*U(2)^3)*p2^3+((-48*p0^2+96*U(2)*p0-48*U(2)^2)*p1^2+96*p0^2-192*U(2)*p0+96*U(2)^2+108*U(1)^2)*p2^2+((12*p0-12*U(2))*p1^4+(48*U(2)-48*p0)*p1^2+108*U(1)*p1+48*p0-48*U(2))*p2-p1^6+6*p1^4+15*p1^2+8)+2*3^(3/2)*U(1)*p2+3^(3/2)*p1)^(2/3)-2*sqrt(3)*p1*(sqrt((64*p0^3-192*U(2)*p0^2+192*U(2)^2*p0-64*U(2)^3)*p2^3+((-48*p0^2+96*U(2)*p0-48*U(2)^2)*p1^2+96*p0^2-192*U(2)*p0+96*U(2)^2+108*U(1)^2)*p2^2+((12*p0-12*U(2))*p1^4+(48*U(2)-48*p0)*p1^2+108*U(1)*p1+48*p0-48*U(2))*p2-p1^6+6*p1^4+15*p1^2+8)+2*3^(3/2)*U(1)*p2+3^(3/2)*p1)^(1/3)+((4*sqrt(3)*i+4)*p0+(-4*sqrt(3)*i-4)*U(2))*p2+(-sqrt(3)*i-1)*p1^2+2*sqrt(3)*i+2)/(4*sqrt(3)*p2*(sqrt((64*p0^3-192*U(2)*p0^2+192*U(2)^2*p0-64*U(2)^3)*p2^3+((-48*p0^2+96*U(2)*p0-48*U(2)^2)*p1^2+96*p0^2-192*U(2)*p0+96*U(2)^2+108*U(1)^2)*p2^2+((12*p0-12*U(2))*p1^4+(48*U(2)-48*p0)*p1^2+108*U(1)*p1+48*p0-48*U(2))*p2-p1^6+6*p1^4+15*p1^2+8)+2*3^(3/2)*U(1)*p2+3^(3/2)*p1)^(1/3));...
                  ((sqrt((64*p0^3-192*U(2)*p0^2+192*U(2)^2*p0-64*U(2)^3)*p2^3+((-48*p0^2+96*U(2)*p0-48*U(2)^2)*p1^2+96*p0^2-192*U(2)*p0+96*U(2)^2+108*U(1)^2)*p2^2+((12*p0-12*U(2))*p1^4+(48*U(2)-48*p0)*p1^2+108*U(1)*p1+48*p0-48*U(2))*p2-p1^6+6*p1^4+15*p1^2+8)+2*3^(3/2)*U(1)*p2+3^(3/2)*p1)^(2/3)-sqrt(3)*p1*(sqrt((64*p0^3-192*U(2)*p0^2+192*U(2)^2*p0-64*U(2)^3)*p2^3+((-48*p0^2+96*U(2)*p0-48*U(2)^2)*p1^2+96*p0^2-192*U(2)*p0+96*U(2)^2+108*U(1)^2)*p2^2+((12*p0-12*U(2))*p1^4+(48*U(2)-48*p0)*p1^2+108*U(1)*p1+48*p0-48*U(2))*p2-p1^6+6*p1^4+15*p1^2+8)+2*3^(3/2)*U(1)*p2+3^(3/2)*p1)^(1/3)+(4*U(2)-4*p0)*p2+p1^2-2)/(2*sqrt(3)*p2*(sqrt((64*p0^3-192*U(2)*p0^2+192*U(2)^2*p0-64*U(2)^3)*p2^3+((-48*p0^2+96*U(2)*p0-48*U(2)^2)*p1^2+96*p0^2-192*U(2)*p0+96*U(2)^2+108*U(1)^2)*p2^2+((12*p0-12*U(2))*p1^4+(48*U(2)-48*p0)*p1^2+108*U(1)*p1+48*p0-48*U(2))*p2-p1^6+6*p1^4+15*p1^2+8)+2*3^(3/2)*U(1)*p2+3^(3/2)*p1)^(1/3))];
            %disp("all xs: "+num2str(xSet(1))+" / "+num2str(xSet(2))+" / "+num2str(xSet(3)));
            %pause; 
            xSet=sortrows([abs(imag(xSet)) real(xSet) imag(xSet)])*[0;1;1i];
            % disp(xSet);
            x = real(xSet(1));
            %assert(isreal(real(xSet(3))),"ERROR: x-coordinate must be real-valued! STOP");
            y = polyval(obj.trackC.poly(secId,:),x); 
            % angle at surface
            gamma = atan(polyval(polyder(obj.trackC.poly(secId,:)),x));
            % left and right end
            X = boundsOfSection(obj, secId);
            % x = q(1); y=q(2)
            %disp("x = "+num2str(x)+"/ X = "+num2str(X(1))+", "+num2str(X(2))+" - sec # = "+ num2str(secId));
            %pause;
            xi = (x-X(1))/(X(2)-X(1));
            
            % now compute eta from the normal vector n and tangent vector t
            t = [1., 2.*p2*x+p1];
            n = [1.,-1./t(2)];
            n = n/sqrt(n*n');
            
            rot = n(2)*t(1)-n(1)-t(2);
            if rot < 0
                n=-1*n;
            end
            DH = U(1)-x;
            DV = U(2)-y;
            if abs(DH)>abs(DV)
                eta = DH/(n(1)*obj.R);
            else
                eta = DV/(n(2)*obj.R);
            end
            
            % Achtung: hier wird der Winkel gegenüber der VERTIKALEN VERMASST!
            % there should be an angle of 90° between orientation and gamma, i.e.
            % orientation-gamma = pi/2
            %orientation = atan2(DV,DH);
            %disp(orientation);
            %ssert(0 < orientation && orientation < pi, "ERROR: angle "+num2str(orientation)+" shouls be positive - this may indicate that the wheel hub is already below the mud .... STOP!");
            % 
        end
                       
        function plotCondition(obj,axle)
            % make a plot of the current situation
            
            U = [obj.U(2*axle+1),obj.U(2*axle+2)]; %#ok<PROPLC>
            
            % wheel
            alpha = -pi:0.1:pi;
            wheel = [U(1)+obj.R*sin(alpha); U(2)-obj.R*cos(alpha)];
                        
            % find relevant sections
            for secId = 1:obj.trackC.nos
                X = obj.boundsOfSection(secId);
                xi = (U(1)-X(1))/(X(2)-X(1));
                if 0<xi && xi<1
                    break;
                end
            end
                       
            % plot with adjecent track sections
            track = [];
            for id = max(1,secId-1):min(obj.trackC.nos,secId+1)
                X = obj.boundsOfSection(id);
                x = X(1):(X(2)-X(1))/20:X(2);
                toadd = [x;polyval(obj.trackC.poly(id,:),x)];
                track = [track, toadd];
            end
            
            hold on;
            %title('current geometry constellation for axle '+num2str(axle));
            %ylabel('h/m # ->'); 
            %xlabel('s/m ->');
            plot(wheel(1,:),wheel(2,:));
            plot(track(1,:),track(2,:));
            
            %pbaspect([0.5 1 1])
            grid on;
    
            hold off;
        end
        
        function contactSpeed(obj)
            % velocities at contact point - using axle speeds V(3...6)
            axle = 1;
            obj.vC(axle,1) = -obj.R*obj.omega+obj.V(4)*obj.singamma(axle)+obj.V(3)*obj.cosgamma(axle);
            obj.vC(axle,2) = obj.V(4)*obj.cosgamma(axle)-obj.V(3)*obj.singamma(axle);

            axle = 2;
            obj.vC(axle,1) = 0.; % not used!
            obj.vC(axle,2) = obj.V(6)*obj.cosgamma(axle)-obj.V(5)*obj.singamma(axle);
      
        end
        
        function [FN, FT] = contactF(obj,axle)
            % returns the value of the force on axle (not an array)
            % as a function of axle (1 or 2) and eta
            % axle is front (1) or rear (2)
            assert(any([1 2] == axle), "ERROR: axle must be front (1) or rear (2)!");

            % Eindringung
            %disp('...........');
            %disp(obj.eta(axle));
            %disp('...........');
            e = -(obj.eta(axle)-1.)*obj.R;
        
            %if e < -obj.eps
            %    FN = 0;
            %    FT = 0;
            %    return;
            %end
            %disp(obj.vC);
            % populate contact speeds vC
            contactSpeed(obj);
            %disp(obj.vC);
            
            FN = obj.springC.getValue(e);
            % include daming
            %FN = FN*exp(-obj.vC(axle,2)/obj.vRef);
            
            % tangential velocity
            vT = obj.tangentialSpeed();       
            FT = FN*obj.muC.getValue(-obj.vC(1,1))-FN*obj.mu*vT/obj.vT0;
        
        end
        
        function vT = tangentialSpeed(obj)
            % returns the speed of the car tangential to the track
            % project speed onto tangential vector (scalar product)
            vT = obj.v(1:2)*[cos(obj.gamma(1));sin(obj.gamma(1))];
        end
    end
end



%xSet = [17.112+3.17467i ; 10.956+2.29358e-15i ; 17.112-3.17467i];
%xSet=sortrows([abs(imag(xSet)) real(xSet) imag(xSet)])*[0;1;1i];
%x = real(xSet(1));
%for i = 3:-1:1
% if abs(imag(xSet(i)))> 1E-9
%     xSet(i)=[]
% end
%end



